"""Background image tilt correction

This example illustrates background tilt correction with qpimage.
In contrast to the 'simple_phase.py' example, the known background data
is not given to the :class:`qpimage.QPImage <qpimage.core.QPImage>`
class. In this particular example, the background tilt correction
achieves an error of about 1% which is sufficient in most quantitative
phase imaging applications.
"""
import matplotlib.pylab as plt
import numpy as np
import qpimage

size = 200
# background phase image with a tilt
bg = np.repeat(np.linspace(0, 1, size), size).reshape(size, size)
bg = .6 * bg - .8 * bg.transpose() + .2
# phase image with random noise
rsobj = np.random.RandomState(47)
phase = rsobj.rand(size, size) - .5 + bg

# create QPImage instance
qpi = qpimage.QPImage(data=phase, which_data="phase")
# compute background with 2d tilt approach
qpi.compute_bg(which_data="phase",  # correct phase image
               fit_offset="fit",  # use bg offset from tilt fit
               fit_profile="tilt",  # perform 2D tilt fit
               border_px=5,  # use 5 px border around image
               )

# plot the properties of `qpi`
fig = plt.figure(figsize=(8, 2.5))
plot_kw = {"vmin": -1,
           "vmax": 1}

ax1 = plt.subplot(131, title="input data")
map1 = ax1.imshow(phase, **plot_kw)
plt.colorbar(map1, ax=ax1, fraction=.046, pad=0.04)

ax2 = plt.subplot(132, title="tilt-corrected")
map2 = ax2.imshow(qpi.pha, **plot_kw)
plt.colorbar(map2, ax=ax2, fraction=.046, pad=0.04)

ax3 = plt.subplot(133, title="tilt error")
map3 = ax3.imshow(bg - qpi.bg_pha)
plt.colorbar(map3, ax=ax3, fraction=.046, pad=0.04)

# disable axes
[ax.axis("off") for ax in [ax1, ax2, ax3]]

plt.tight_layout(pad=0, h_pad=0, w_pad=0)
plt.show()



import numpy as np

import qpimage


def test_poly2o():
    """simple 2nd order polynomial test
    """
    size = 40
    mx = .5
    my = -.3
    mxy = .1
    ax = -.05
    ay = .04
    x = np.linspace(-10, 10, size).reshape(-1, 1)
    y = np.linspace(-10, 10, size).reshape(1, -1)
    off = 1

    phase = off \
        + ax * x ** 2 \
        + ay * y ** 2 \
        + mx * x \
        + my * y \
        + mxy * x * y

    qpi = qpimage.QPImage(data=phase, which_data="phase", h5dtype="float64")
    qpi.compute_bg(which_data="phase",
                   fit_profile="poly2o",
                   from_mask=np.ones_like(phase, dtype=bool))

    assert not np.allclose(phase, 0, atol=1e-14, rtol=0)
    assert np.allclose(qpi.pha, 0, atol=1e-14, rtol=0)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Simple phase

This example illustrates the simple usage of the
:class:`qpimage.QPImage <qpimage.core.QPImage>` class for reading and
managing quantitative phase data. The attribute ``QPImage.pha`` yields
the background-corrected phase data and the attribute
``QPImage.bg_pha`` yields the background phase image.
"""
import matplotlib.pylab as plt
import numpy as np
import qpimage

size = 200
# background phase image with a tilt
bg = np.repeat(np.linspace(0, 1, size), size).reshape(size, size)
# phase image with random noise
phase = np.random.rand(size, size) + bg

# create QPImage instance
qpi = qpimage.QPImage(data=phase, bg_data=bg, which_data="phase")

# plot the properties of `qpi`
plt.figure(figsize=(8, 3))
plot_kw = {"vmin": -1,
           "vmax": 2}

plt.subplot(131, title="fake input phase")
plt.imshow(phase, **plot_kw)

plt.subplot(132, title="fake background phase")
plt.imshow(qpi.bg_pha, **plot_kw)

plt.subplot(133, title="corrected phase")
plt.imshow(qpi.pha, **plot_kw)

plt.tight_layout()
plt.show()



import h5py

from .core import QPImage
from .meta import MetaDict


class QPSeries(object):
    _instances = 0

    def __init__(self, qpimage_list=[], meta_data={},
                 h5file=None, h5mode="a", identifier=None):
        """Quantitative phase image series

        Parameters
        ----------
        qpimage_list: list
            A list of instances of :class:`qpimage.QPImage`.
        meta_data: dict
            Metadata associated with the input data
            (see :const:`qpimage.META_KEYS`). This overrides
            the metadata of the QPImages in `qpimage_list` and, if
            `h5file` is given and `h5mode` is not "r", overrides
            the metadata in `h5file`.
        h5file: str, h5py.Group, h5py.File, or None
            A path to an HDF5 data file where all data is cached. If
            set to `None` (default), all data will be handled in
            memory using the "core" driver of the :mod:`h5py`'s
            :class:`h5py:File` class. If the file does not exist,
            it is created. If the file already exists, it is opened
            with the file mode defined by `hdf5_mode`. If this is
            an instance of h5py.Group or h5py.File, then this will
            be used to internally store all data. If `h5file` is given
            and `qpimage_list` is not empty, all QPImages in
            `qpimage_list` are appended to `h5file` in the given order.
        h5mode: str
            Valid file modes are (only applies if `h5file` is a path):

            - "r": Readonly, file must exist
            - "r+": Read/write, file must exist
            - "w": Create file, truncate if exists
            - "w-" or "x": Create file, fail if exists
            - "a": Read/write if exists, create otherwise (default)
        """
        if qpimage_list and not isinstance(qpimage_list, list):
            msg = "`qpimage_list` must be a list!"
            if isinstance(qpimage_list, str):
                msg += " Did you mean `h5file={}`?".format(qpimage_list)
            raise ValueError(msg)
        if isinstance(h5file, h5py.Group):
            self.h5 = h5file
            self._do_h5_cleanup = False
        else:
            if h5file is None:
                h5kwargs = {"name": "qpseries{}.h5".format(
                    QPSeries._instances),
                    "driver": "core",
                    "backing_store": False,
                    "mode": "w"}
            else:
                h5kwargs = {"name": h5file,
                            "mode": h5mode}
            self.h5 = h5py.File(**h5kwargs)
            self._do_h5_cleanup = True
        QPSeries._instances += 1

        if meta_data and h5mode == "r":
            msg = "`h5mode` must not be 'r' if `meta_data` is given!"
            raise ValueError(msg)

        # make sure self.h5 is not itself a QPImage file
        if "phase" in self.h5 and "amplitude" in self.h5:
            raise ValueError(
                "`h5file` is a QPImage file, not a QPSeries file!")

        # Write QPimage data to h5 file
        for qpi in qpimage_list:
            self.add_qpimage(qpi)

        # Update meta data
        if meta_data:
            meta = MetaDict(meta_data)
            for ii in range(len(self)):
                qpii = self.get_qpimage(index=ii)
                for mk in meta:
                    qpii.h5.attrs[mk] = meta[mk]

        # Set identifier
        if identifier:
            self.h5.attrs["identifier"] = identifier

    def __contains__(self, qpid):
        """test whether a QPImage with the given identifier exists"""
        for ii in range(len(self)):
            qpi = self[ii]
            if "identifier" in qpi and qpi["identifier"] == qpid:
                exists = True
                break
        else:
            exists = False
        return exists

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._do_h5_cleanup:
            self.h5.flush()
            self.h5.close()

    def __getitem__(self, index):
        return self.get_qpimage(index)

    def __iter__(self):
        for ii in range(len(self)):
            yield self[ii]

    def __len__(self):
        keys = list(self.h5.keys())
        keys = [kk for kk in keys if kk.startswith("qpi_")]
        return len(keys)

    @property
    def identifier(self):
        """unique identifier of the series"""
        if "identifier" in self.h5.attrs:
            return self.h5.attrs["identifier"]
        else:
            return None

    def add_qpimage(self, qpi, identifier=None, bg_from_idx=None):
        """Add a QPImage instance to the QPSeries

        Parameters
        ----------
        qpi: qpimage.QPImage
            The QPImage that is added to the series
        identifier: str
            Identifier key for `qpi`
        bg_from_idx: int or None
            Use the background data from the data stored in this index,
            creating hard links within the HDF5 file.
            (Saves memory if e.g. all qpimages is corrected with the same data)
        """
        if not isinstance(qpi, QPImage):
            raise ValueError("`fli` must be instance of QPImage!")
        if "identifier" in qpi and identifier is None:
            identifier = qpi["identifier"]
        if identifier and identifier in self:
            msg = "The identifier '{}' already ".format(identifier) \
                  + "exists! You can either change the identifier of " \
                  + " '{}' or remove it.".format(qpi)
            raise ValueError(msg)
        # determine number of qpimages
        num = len(self)
        # indices start at zero; do not add 1
        name = "qpi_{}".format(num)
        group = self.h5.create_group(name)
        thisqpi = qpi.copy(h5file=group)

        if bg_from_idx is not None:
            # Create hard links
            refqpi = self[bg_from_idx]
            thisqpi._amp.set_bg(bg=refqpi._amp.h5["bg_data"]["data"])
            thisqpi._pha.set_bg(bg=refqpi._pha.h5["bg_data"]["data"])

        if identifier:
            # set identifier
            group.attrs["identifier"] = identifier

    def get_qpimage(self, index):
        """Return a single QPImage of the series

        Parameters
        ----------
        index: int or str
            Index or identifier of the QPImage

        Notes
        -----
        Instead of ``qps.get_qpimage(index)``, it is possible
        to use the shorthand ``qps[index]``.
        """
        if isinstance(index, str):
            # search for the identifier
            for ii in range(len(self)):
                qpi = self[ii]
                if "identifier" in qpi and qpi["identifier"] == index:
                    group = self.h5["qpi_{}".format(ii)]
                    break
            else:
                msg = "QPImage identifier '{}' not found!".format(index)
                raise KeyError(msg)
        else:
            # integer index
            if index < -len(self):
                msg = "Index {} out of bounds for QPSeries of size {}!".format(
                    index, len(self))
                raise ValueError(msg)
            elif index < 0:
                index += len(self)
            name = "qpi_{}".format(index)
            if name in self.h5:
                group = self.h5[name]
            else:
                msg = "Index {} not found for QPSeries of length {}".format(
                    index, len(self))
                raise KeyError(msg)
        return QPImage(h5file=group)



import numpy as np
import pytest
import qpimage


def test_qpimage_qpretrieve_oah_deprecated():
    # create fake hologram
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    kx = -.6
    ky = -.4
    disk_max = 1.5
    # there is a phase disk as data in the hologram
    data = disk_max * ((x - size / 2)**2 + (y - size / 2)**2 < 30**2)
    image = np.sin(kx * x + ky * y + data)
    with pytest.warns(DeprecationWarning, match="qpretrieve_kw"):
        qpi = qpimage.QPImage(image,
                              which_data="raw-oah",
                              holo_kw={"filter_name": "gauss"})
        qpi.compute_bg(which_data="phase",
                       fit_offset="fit",
                       fit_profile="tilt",
                       border_px=5)
    assert np.allclose(disk_max, qpi.pha.max(), rtol=.01, atol=0)


def test_qpimage_qpretrieve_oah():
    # create fake hologram
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    kx = -.6
    ky = -.4
    disk_max = 1.5
    # there is a phase disk as data in the hologram
    data = disk_max * ((x - size / 2)**2 + (y - size / 2)**2 < 30**2)
    image = np.sin(kx * x + ky * y + data)
    qpi = qpimage.QPImage(image,
                          which_data="raw-oah",
                          qpretrieve_kw={"filter_name": "gauss"})
    qpi.compute_bg(which_data="phase",
                   fit_offset="fit",
                   fit_profile="tilt",
                   border_px=5)
    assert np.allclose(disk_max, qpi.pha.max(), rtol=.01, atol=0)



"""Background image 2nd order polynomial correction

This example extends the tilt correction ('background_tilt.py') to
a second order polynomial correction for samples that exhibit
more sophisticated phase aberrations. The phase background correction
is computed from a ten pixel wide frame around the image.
The phase data shown are computed from a hologram of a single
myeloid leukemia cell (HL60) recorded using digital holographic
microscopy (DHM) (see :cite:`Schuermann2015`).
"""
import matplotlib.pylab as plt
import numpy as np
# The data are stored in a .jpg file (lossy compression).
# If `PIL` is not found, try installing the `pillow` package.
from PIL import Image
import qpimage

edata = np.array(Image.open("./data/hologram_cell_curved_bg.jpg"))

# create QPImage instance
qpi = qpimage.QPImage(data=edata, which_data="raw-oah")
pha0 = qpi.pha

# background correction using tilt only
qpi.compute_bg(which_data=["phase"],
               fit_offset="fit",
               fit_profile="tilt",
               border_px=10,
               )
pha_tilt = qpi.pha

# background correction using polynomial
qpi.compute_bg(which_data=["phase"],
               fit_offset="fit",
               fit_profile="poly2o",
               border_px=10,
               )
pha_poly2o = qpi.pha

# plot phase data
fig = plt.figure(figsize=(8, 3.3))

phakw = {"cmap": "viridis",
         "interpolation": "bicubic",
         "vmin": pha_poly2o.min(),
         "vmax": pha_poly2o.max()}

ax1 = plt.subplot(131, title="input phase")
map1 = ax1.imshow(pha0, **phakw)
plt.colorbar(map1, ax=ax1, fraction=.067, pad=0.04)

ax2 = plt.subplot(132, title="tilt correction only")
map2 = ax2.imshow(pha_tilt, **phakw)
plt.colorbar(map2, ax=ax2, fraction=.067, pad=0.04)

ax3 = plt.subplot(133, title="polynomial correction")
map3 = ax3.imshow(pha_poly2o, **phakw)
plt.colorbar(map3, ax=ax3, fraction=.067, pad=0.04)

# disable axes
[ax.axis("off") for ax in [ax1, ax2, ax3]]

plt.tight_layout(w_pad=0)
plt.show()



import os
import tempfile

import numpy as np

import qpimage


def test_phase_array():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    with qpimage.QPImage(phase, which_data="phase", h5dtype="float64") as qpi:
        assert np.all(qpi.pha == phase)


def test_file():
    h5file = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    # Write data to disk
    with qpimage.QPImage(phase,
                         which_data="phase",
                         h5file=h5file,
                         h5mode="a",
                         ) as qpi:
        p1 = qpi.pha
        a1 = qpi.amp
    # Open data read-only
    qpi2 = qpimage.QPImage(h5file=h5file, h5mode="r")
    assert np.all(p1 == qpi2.pha)
    assert np.all(a1 == qpi2.amp)
    # cleanup
    try:
        os.remove(h5file)
    except OSError:
        pass


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import numpy as np

import qpimage
from qpimage import meta
from qpimage import bg_estimate


def test_bg_estimate_errors():
    data = np.zeros((5, 5))

    try:
        bg_estimate.estimate(data, fit_offset="unknown")
    except ValueError:
        pass
    else:
        assert False, "unknown is not a valid key"

    try:
        bg_estimate.estimate(data, fit_profile="unknown")
    except ValueError:
        pass
    else:
        assert False, "unknown is not a valid key"

    try:
        bg_estimate.estimate(data,
                             fit_offset="fit",
                             fit_profile="offset")
    except ValueError:
        pass
    else:
        assert False, "offset cannot be fitted"


def test_mask():
    size = 200
    data = np.zeros((size, size), dtype=float)
    data[size // 2:] = 1
    mask = data == 0
    qpi = qpimage.QPImage(data=data, which_data="phase")
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   from_mask=mask,
                   )
    assert np.allclose(qpi.pha, data)
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   from_mask=~mask,
                   )
    assert np.allclose(qpi.pha, data - 1)


def test_border_m():
    size = 200
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=.4, scale=.1, size=(size, size))
    pixel_size = .1e-6  # 1px = .1µm
    border_px = 5
    border_m = border_px * pixel_size

    qpi = qpimage.QPImage(data=data, which_data="phase")
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_px=5)
    data_px = qpi.pha
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_px=6)
    data_px_false = qpi.pha
    # Make sure different borders yield different results
    assert not np.all(data_px == data_px_false)
    # Make sure QPImage needs the "pixel size" meta data
    try:
        qpi.compute_bg(which_data="phase",
                       fit_offset="mean",
                       fit_profile="offset",
                       border_m=border_m)
    except meta.MetaDataMissingError:
        pass
    else:
        assert False, "meta data required"
    # Create an instance with the correct pixel size
    qpi2 = qpimage.QPImage(data=data, which_data="phase",
                           meta_data={"pixel size": pixel_size})
    qpi2.compute_bg(which_data="phase",
                    fit_offset="mean",
                    fit_profile="offset",
                    border_m=border_m)
    assert np.all(data_px == qpi2.pha)
    # Also test with rounding
    qpi2.compute_bg(which_data="phase",
                    fit_offset="mean",
                    fit_profile="offset",
                    border_m=border_m * 1.05)
    assert np.all(data_px == qpi2.pha)
    qpi2.compute_bg(which_data="phase",
                    fit_offset="mean",
                    fit_profile="offset",
                    border_m=border_m * .95)
    assert np.all(data_px == qpi2.pha)


def test_border_perc():
    size = 200
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=.4, scale=.1, size=(size, size))

    qpi = qpimage.QPImage(data=data, which_data="phase")
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_px=5)
    data_px = qpi.pha
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_perc=2.5)
    assert np.all(qpi.pha == data_px)
    # rounding
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_perc=2.4)
    assert np.all(qpi.pha == data_px)
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_perc=2.1)  # 4px
    assert not np.all(qpi.pha == data_px)


def test_get_mask():
    size = 200
    data = np.zeros((size, size), dtype=float)
    mask = np.zeros_like(data, dtype=bool)
    mask[::2, ::2] = True
    qpi = qpimage.QPImage(data=data, which_data="phase")
    # only mask
    bin1 = qpi.compute_bg(which_data="phase",
                          fit_offset="mean",
                          fit_profile="offset",
                          from_mask=mask,
                          ret_mask=True)
    assert np.all(mask == bin1)
    # mask with border
    bin2 = qpi.compute_bg(which_data="phase",
                          fit_offset="mean",
                          fit_profile="offset",
                          from_mask=mask,
                          border_px=5,
                          ret_mask=True)
    mask2 = mask.copy()
    mask2[5:-5, 5:-5] = False
    assert np.all(mask2 == bin2)


def test_tilt():
    size = 100
    data = np.zeros((size, size))
    data += np.linspace(0, .1, size).reshape(-1, 1)
    bg = bg_estimate.estimate(data=data,
                              fit_offset="fit",
                              fit_profile="tilt",
                              border_px=1)
    assert np.allclose(data, bg)



import qpimage
import qpimage.meta


def test_meta():
    mm = qpimage.meta.MetaDict()
    mm["wavelength"] = 100e-9
    assert mm["wavelength"] == 100e-9


def test_meta_error():
    mm = qpimage.meta.MetaDict()
    try:
        mm["peter"] = "hans"
    except KeyError:
        pass
    else:
        assert False, "invalid key 'peter' should not work!"

    try:
        qpimage.meta.MetaDict({"peter2": "hans2"})
    except KeyError:
        pass
    else:
        assert False, "invalid key 'peter2' should not work!"

    try:
        mm["peter"]
    except KeyError:
        pass
    else:
        assert False, "invalid key 'peter' should raise KeyError!"

    try:
        mm["medium index"]
    except qpimage.meta.MetaDataMissingError:
        pass
    else:
        assert False, "valid key undefined should raise MetaDataMissingError"


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import numpy as np

import qpimage


def test_offset_mean_simple():
    size = 200
    bg = 2.5
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=bg, scale=.1, size=(size, size))
    assert np.allclose(np.mean(data), 2.5003227280152887)
    qpi = qpimage.QPImage(data=data, which_data="phase", h5dtype="float64")
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_px=5)
    assert np.allclose(np.mean(qpi.pha), -0.0011984843060164104)


def test_offset_mode_simple():
    size = 200
    bg = 2.5
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=bg, scale=.1, size=(size, size))
    assert np.allclose(np.mean(data), 2.5003227280152887)
    qpi = qpimage.QPImage(data=data, which_data="phase", h5dtype="float64")
    qpi.compute_bg(which_data="phase",
                   fit_offset="mode",
                   fit_profile="offset",
                   border_px=5)
    # Not so good result because of poor statistics
    assert np.allclose(np.mean(qpi.pha), -0.044022731691440088)


def test_offset_gauss_simple():
    size = 200
    bg = 2.5
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=bg, scale=.1, size=(size, size))
    assert np.allclose(np.mean(data), 2.5003227280152887)
    qpi = qpimage.QPImage(data=data, which_data="phase", h5dtype="float64")
    qpi.compute_bg(which_data="phase",
                   fit_offset="gauss",
                   fit_profile="offset",
                   border_px=5)
    # This yields the best results, because the input noise
    # is gaussian.
    assert np.allclose(np.average(qpi.pha), -0.00039833492077054762)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import lmfit
import numpy as np


#: valid values for keyword argument `fit_offset` in :func:`estimate`
VALID_FIT_OFFSETS = ["fit", "gauss", "mean", "mode"]
#: valid values for keyword argument `fit_profile` in :func:`estimate`
VALID_FIT_PROFILES = ["offset", "poly2o", "tilt"]


def estimate(data, fit_offset="mean", fit_profile="tilt",
             border_px=0, from_mask=None, ret_mask=False):
    """Estimate the background value of an image

    Parameters
    ----------
    data: np.ndarray
        Data from which to compute the background value
    fit_profile: str
        The type of background profile to fit:

        - "offset": offset only
        - "poly2o": 2D 2nd order polynomial with mixed terms
        - "tilt": 2D linear tilt with offset (default)
    fit_offset: str
        The method for computing the profile offset

        - "fit": offset as fitting parameter
        - "gauss": center of a gaussian fit
        - "mean": simple average
        - "mode": mode (see `qpimage.bg_estimate.mode`)
    border_px: float
        Assume that a frame of `border_px` pixels around
        the image is background.
    from_mask: boolean np.ndarray or None
        Use a boolean array to define the background area.
        The boolean mask must have the same shape as the
        input data. `True` elements are used for background
        estimation.
    ret_mask: bool
        Return the boolean mask used to compute the background.

    Notes
    -----
    If both `border_px` and `from_mask` are given, the
    intersection of the two is used, i.e. the positions
    where both, the frame mask and `from_mask`, are
    `True`.
    """
    if fit_profile not in VALID_FIT_PROFILES:
        msg = "`fit_profile` must be one of {}, got '{}'".format(
            VALID_FIT_PROFILES,
            fit_profile)
        raise ValueError(msg)
    if fit_offset not in VALID_FIT_OFFSETS:
        msg = "`fit_offset` must be one of {}, got '{}'".format(
            VALID_FIT_OFFSETS,
            fit_offset)
        raise ValueError(msg)
    # initial mask image
    if from_mask is not None:
        assert isinstance(from_mask, np.ndarray)
        mask = from_mask.copy()
    else:
        mask = np.ones_like(data, dtype=bool)
    # multiply with border mask image (intersection)
    if border_px > 0:
        border_px = int(np.round(border_px))
        mask_px = np.zeros_like(mask)
        mask_px[:border_px, :] = True
        mask_px[-border_px:, :] = True
        mask_px[:, :border_px] = True
        mask_px[:, -border_px:] = True
        # intersection
        np.logical_and(mask, mask_px, out=mask)
    # compute background image
    if fit_profile == "tilt":
        bgimg = profile_tilt(data, mask)
    elif fit_profile == "poly2o":
        bgimg = profile_poly2o(data, mask)
    else:
        bgimg = np.zeros_like(data, dtype=float)
    # add offsets
    if fit_offset == "fit":
        if fit_profile == "offset":
            msg = "`fit_offset=='fit'` only valid when `fit_profile!='offset`"
            raise ValueError(msg)
        # nothing else to do here, using offset from fit
    elif fit_offset == "gauss":
        bgimg += offset_gaussian((data - bgimg)[mask])
    elif fit_offset == "mean":
        bgimg += np.mean((data - bgimg)[mask])
    elif fit_offset == "mode":
        bgimg += offset_mode((data - bgimg)[mask])

    if ret_mask:
        ret = (bgimg, mask)
    else:
        ret = bgimg
    return ret


def offset_gaussian(data):
    """Fit a gaussian model to `data` and return its center"""
    nbins = 2 * int(np.ceil(np.sqrt(data.size)))
    mind, maxd = data.min(), data.max()
    drange = (mind - (maxd - mind) / 2, maxd + (maxd - mind) / 2)
    histo = np.histogram(data, nbins, density=True, range=drange)
    dx = abs(histo[1][1] - histo[1][2]) / 2
    hx = histo[1][1:] - dx
    hy = histo[0]
    # fit gaussian
    gauss = lmfit.models.GaussianModel()
    pars = gauss.guess(hy, x=hx)
    out = gauss.fit(hy, pars, x=hx)
    return out.params["center"]


def offset_mode(data):
    """Compute Mode using a histogram with `sqrt(data.size)` bins"""
    nbins = int(np.ceil(np.sqrt(data.size)))
    mind, maxd = data.min(), data.max()
    histo = np.histogram(data, nbins, density=True, range=(mind, maxd))
    dx = abs(histo[1][1] - histo[1][2]) / 2
    hx = histo[1][1:] - dx
    hy = histo[0]
    idmax = np.argmax(hy)
    return hx[idmax]


def profile_tilt(data, mask):
    """Fit a 2D tilt to `data[mask]`"""
    params = lmfit.Parameters()
    params.add(name="mx", value=0)
    params.add(name="my", value=0)
    params.add(name="off", value=np.average(data[mask]))
    fr = lmfit.minimize(tilt_residual, params, args=(data, mask))
    bg = tilt_model(fr.params, data.shape)
    return bg


def profile_poly2o(data, mask):
    """Fit a 2D 2nd order polynomial to `data[mask]`"""
    # lmfit
    params = lmfit.Parameters()
    params.add(name="mx", value=0)
    params.add(name="my", value=0)
    params.add(name="mxy", value=0)
    params.add(name="ax", value=0)
    params.add(name="ay", value=0)
    params.add(name="off", value=np.average(data[mask]))
    fr = lmfit.minimize(poly2o_residual, params, args=(data, mask))
    bg = poly2o_model(fr.params, data.shape)
    return bg


def poly2o_model(params, shape):
    """lmfit 2nd order polynomial model"""
    mx = params["mx"].value
    my = params["my"].value
    mxy = params["mxy"].value
    ax = params["ax"].value
    ay = params["ay"].value
    off = params["off"].value
    bg = np.zeros(shape, dtype=float) + off
    x = np.arange(bg.shape[0]) - bg.shape[0] // 2
    y = np.arange(bg.shape[1]) - bg.shape[1] // 2
    x = x.reshape(-1, 1)
    y = y.reshape(1, -1)
    bg += ax * x**2 + ay * y**2 + mx * x + my * y + mxy * x * y
    return bg


def poly2o_residual(params, data, mask):
    """lmfit 2nd order polynomial residuals"""
    bg = poly2o_model(params, shape=data.shape)
    res = (data - bg)[mask]
    return res.flatten()


def tilt_model(params, shape):
    """lmfit tilt model"""
    mx = params["mx"].value
    my = params["my"].value
    off = params["off"].value
    bg = np.zeros(shape, dtype=float) + off
    x = np.arange(bg.shape[0]) - bg.shape[0] // 2
    y = np.arange(bg.shape[1]) - bg.shape[1] // 2
    x = x.reshape(-1, 1)
    y = y.reshape(1, -1)
    bg += mx * x + my * y
    return bg


def tilt_residual(params, data, mask):
    """lmfit tilt residuals"""
    bg = tilt_model(params, shape=data.shape)
    res = (data - bg)[mask]
    return res.flatten()



import pathlib
import warnings

import h5py
import nrefocus
import numpy as np
import qpretrieve
from skimage.restoration import unwrap_phase

from .image_data import Amplitude, Phase, write_image_dataset
from .meta import MetaDict, DATA_KEYS, META_KEYS
from ._version import version as __version__

#: valid combinations for keyword argument `which_data`
VALID_INPUT_DATA = ["field",
                    "hologram",  # deprecated
                    "raw-oah",  # off-axis holography
                    "raw-qlsi",  # quadri-wave lateral shearing interferometry
                    "phase",
                    ("phase", "amplitude"),
                    ("phase", "intensity"),
                    ]


class QPImage(object):
    # required to create in-memory HDF5 files with unique fd
    _instances = 0

    def __init__(self, data=None, bg_data=None, which_data="phase",
                 meta_data=None, qpretrieve_kw=None, holo_kw=None,
                 proc_phase=True, h5file=None, h5mode="a", h5dtype="float32"):
        """Quantitative phase image manipulation

        This class implements various tasks for quantitative phase
        imaging, including phase unwrapping, background correction,
        numerical focusing, and data export.

        Parameters
        ----------
        data: 2d ndarray (float or complex) or list
            The experimental data (see `which_data`)
        bg_data: 2d ndarray (float or complex), list, or `None`
            The background data (must be same type as `data`)
        which_data: str or tuple
            String or comma-separated list of strings indicating
            the order and type of input data. Valid values are
            defined in :const:`VALID_INPUT_DATA`, where phase and
            amplitude/intensity data require an indexable object
            with the phase data as first element.
        meta_data: dict or qpimage.MetaDict
            Metadata associated with the input data.
            see :data:`qpimage.meta.META_KEYS`
        qpretrieve_kw: dict
            Keyword arguments passed to
            :ref:`qpretrieve <qpretrieve:index>` for
            phase retrieval from interferometric data.
        holo_kw: dict
            This is deprecated, please use `qpretrieve_kw` instead.
            Special keyword arguments for phase retrieval from
            hologram data (`which_data="raw-oah"`).

            .. versionadded:: 0.1.6
        proc_phase: bool
            Process the phase data. This includes phase unwrapping
            using :func:`skimage.restoration.unwrap_phase` and
            correcting for 2PI phase offsets (The offset is estimated
            from a 1px-wide border around the image).

            .. versionadded:: 0.6.0
                Previous versions always performed phase unwrapping
                and did so without offset correction
        h5file: str, pathlib.Path, h5py.Group, h5py.File, or None
            A path to an HDF5 data file where all data is cached. If
            set to `None` (default), all data will be handled in
            memory using the "core" driver of the :mod:`h5py`'s
            :class:`h5py:File` class. If the file does not exist,
            it is created. If the file already exists, it is opened
            with the file mode defined by `hdf5_mode`. If this is
            an instance of h5py.Group or h5py.File, then this will
            be used to internally store all data.
        h5mode: str
            Valid file modes are (only applies if `h5file` is a path)

            - "r": Readonly, file must exist
            - "r+": Read/write, file must exist
            - "w": Create file, truncate if exists
            - "w-" or "x": Create file, fail if exists
            - "a": Read/write if exists, create otherwise (default)

        h5dtype: str
            The datatype in which to store the image data. The default
            is "float32" which is sufficient for 2D image analysis and
            consumes only half the disk space of the numpy default
            "float64".

        Notes
        -----
        QPImage is slicable; the following returns a new QPImage with
        the same metadata, but with all background corrections merged.

        .. code-block:: python

            qpi = QPImage(data=...)
            qpi_scliced = qpi[10:20, 40:30]
        """
        if qpretrieve_kw is None:
            qpretrieve_kw = {}
        if holo_kw is not None:
            warnings.warn(
                "`holo_kw` is deprecated! Please use `qpretrieve_kw` instead",
                DeprecationWarning)
            # map deprecated parameters to `qpretrieve_kw`
            for key in holo_kw:
                if key == "sideband":
                    if holo_kw[key] in [-1, 1]:
                        qpretrieve_kw["invert_phase"] = holo_kw[key] == -1
                    else:
                        qpretrieve_kw["sideband_freq"] = tuple(holo_kw[key])
                        qpretrieve_kw["invert_phase"] = False
                elif key == "zero_pad":
                    qpretrieve_kw["padding"] = holo_kw["zero_pad"]
                else:
                    qpretrieve_kw[key] = holo_kw[key]

        if meta_data is None:
            meta_data = {}
        if (data is not None and
                not isinstance(data, (np.ndarray, list, tuple))):
            msg = "`data` must be numpy.ndarray!"
            if isinstance(data, (str, pathlib.Path)):
                msg += " Did you mean `h5file={}`?".format(data)
            raise ValueError(msg)
        if isinstance(h5file, h5py.Group):
            self.h5 = h5file
            self._do_h5_cleanup = False
        else:
            if h5file is None:
                h5kwargs = {"name": "qpimage{}.h5".format(QPImage._instances),
                            "driver": "core",
                            "backing_store": False,
                            "mode": "w"}
            else:
                h5kwargs = {"name": h5file,
                            "mode": h5mode}
            self.h5 = h5py.File(**h5kwargs)
            self._do_h5_cleanup = True
        QPImage._instances += 1
        #: hologram processing keyword arguments
        self.qpretrieve_kw = qpretrieve_kw
        # set meta data
        meta = MetaDict(meta_data)
        for key in meta:
            self.h5.attrs[key] = meta[key]
        if "qpimage version" not in self.h5.attrs:
            self.h5.attrs["qpimage version"] = __version__
        # set data
        self._amp = Amplitude(self.h5.require_group("amplitude"),
                              h5dtype=h5dtype)
        self._pha = Phase(self.h5.require_group("phase"),
                          h5dtype=h5dtype)
        if data is not None:
            # Compute phase and amplitude from input data.
            # Note that for QLSI data, the background correction step
            # takes place in the gradient during wavefront computation.
            # Here we allow this to happen and only set the background
            # data if there are any.
            amp, pha, bg_amp, bg_pha = self._get_amp_pha(
                data=data,
                bg_data=bg_data,
                which_data=which_data,
                proc_phase=proc_phase)
            self._amp["raw"] = amp
            self._pha["raw"] = pha
            # set background data
            if bg_pha is not None:
                self.set_bg_data(bg_data=(bg_pha, bg_amp),
                                 which_data="phase,amplitude")
        self.h5dtype = h5dtype
        # :mod:`nrefocus` interface class
        self._refocuser = None

    def __enter__(self):
        return self

    def __eq__(self, other):
        datame = [self.meta[k] for k in self.meta if k in DATA_KEYS]
        dataot = [other.meta[k] for k in other.meta if k in DATA_KEYS]

        if (isinstance(other, QPImage) and
            self.shape == other.shape and
            np.allclose(self.amp, other.amp) and
            np.allclose(self.pha, other.pha) and
                datame == dataot):
            return True
        else:
            return False

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._do_h5_cleanup:
            self.h5.flush()
            self.h5.close()

    def __contains__(self, key):
        return key in self.h5.attrs

    def __getitem__(self, given):
        """Slice QPImage `pha` and `amp` and return a new QPImage

        The background data of the returned QPImage is merged into
        the "data" background array, i.e. there will be no "fit"
        background array.
        """
        if isinstance(given, (slice, tuple)):
            # return new QPImage
            qpi = QPImage(data=(self.raw_pha[given], self.raw_amp[given]),
                          bg_data=(self.bg_pha[given], self.bg_amp[given]),
                          which_data=("phase", "amplitude"),
                          meta_data=self.meta,
                          proc_phase=False)
            return qpi
        elif isinstance(given, str):
            # return meta data
            return self.meta[given]
        else:
            msg = "Only slicing and meta data keys allowed for `__getitem__`"
            raise ValueError(msg)

    def __repr__(self):
        if "identifier" in self:
            ident = self["identifier"]
        else:
            ident = hex(id(self))
        rep = "QPImage <{}>, {x}x{y}px".format(ident,
                                               x=self._amp.raw.shape[0],
                                               y=self._amp.raw.shape[1],
                                               )
        if "wavelength" in self:
            wl = self["wavelength"]
            if 2000e-9 > wl > 10e-9:
                # convenience for light microscopy
                rep += ", λ={:.1f}nm".format(wl * 1e9)
            else:
                rep += ", λ={:.2e}m".format(wl)

        return rep

    def __setitem__(self, key, value):
        if key not in META_KEYS:
            raise KeyError("Unknown meta data key: {}".format(key))
        else:
            self.h5.attrs[key] = value

    @staticmethod
    def _conv_which_data(which_data):
        """Convert which data to string or tuple

        This function improves user convenience,
        as `which_data` may be of several types
        (str, ,str with spaces and commas, list, tuple) which
        is internally handled by this method.
        """
        if isinstance(which_data, str):
            which_data = which_data.lower().strip()
            if which_data.count(","):
                # convert comma string to list
                which_data = [w.strip() for w in which_data.split(",")]
                # remove empty strings
                which_data = [w for w in which_data if w]
                if len(which_data) == 1:
                    return which_data[0]
                else:
                    # convert to tuple
                    return tuple(which_data)
            else:
                return which_data
        elif isinstance(which_data, (list, tuple)):
            which_data = [w.lower().strip() for w in which_data]
            return tuple(which_data)
        elif which_data is None:
            return None
        else:
            msg = "unknown type for `which_data`: {}".format(which_data)
            raise ValueError(msg)

    def _get_amp_pha(self, data, which_data, bg_data=None, proc_phase=True):
        """Convert input data to phase and amplitude

        Parameters
        ----------
        data: 2d ndarray (float or complex) or list
            The experimental data (see `which_data`)
        which_data: str
            String or comma-separated list of strings indicating
            the order and type of input data. Valid values are
            defined in :const:`VALID_INPUT_DATA`, where phase and
            amplitude/intensity data require an indexable object
            with the phase data as first element.
        proc_phase: bool
            Process the phase data. This includes phase unwrapping
            using :func:`skimage.restoration.unwrap_phase` and
            correcting for 2PI phase offsets (The offset is estimated
            from a 1px-wide border around the image).

            .. versionadded:: 0.6.0
                Previous versions always performed phase unwrapping
                and did so without offset correction

        Returns
        -------
        amp, pha: .Amplitdue, .Phase
            amplitude and phase retrieved from the input data
        bg_amp, bg_pha: .Amplitdue, .Phase or None, None
            background amplitude and phase retrieved from the input
            data (if applicable)
        """
        which_data = QPImage._conv_which_data(which_data)

        if which_data == "field":
            amp = np.abs(data)
            pha = np.angle(data)
        elif which_data == "phase":
            amp = np.ones_like(data)
            pha = data
        elif which_data == ("phase", "amplitude"):
            amp = data[1]
            pha = data[0]
        elif which_data == ("phase", "intensity"):
            amp = np.sqrt(data[1])
            pha = data[0]
        elif which_data in ["hologram", "raw-oah"]:
            if which_data == "hologram":
                warnings.warn("The 'hologram' data type is deprecated, "
                              + "please use 'raw-oah' instead!",
                              DeprecationWarning)
            oah = qpretrieve.OffAxisHologram(data=data, **self.qpretrieve_kw)
            amp, pha, _, _ = self._get_amp_pha(oah.run_pipeline(),
                                               which_data="field")
        elif which_data == "raw-qlsi":
            qlsi = qpretrieve.QLSInterferogram(
                data=data,
                reference=bg_data,
                **self.qpretrieve_kw)
            qlsi.run_pipeline()
            amp = qlsi.amplitude
            pha = qlsi.phase
            # Set bg_data to None so that we will not process it separately
            bg_data = None
        else:
            raise ValueError(
                f"`which_data` must be one of {VALID_INPUT_DATA}!")

        if amp.size == 0 or pha.size == 0:
            msg = "`data` with shape {} has zero size!".format(amp.shape)
            raise ValueError(msg)
        if proc_phase:
            # phase unwrapping (take into account nans)
            nanmask = np.isnan(pha)
            if np.sum(nanmask):
                # create masked array
                # skimage.restoration.unwrap_phase cannot handle nan data
                # (even if masked)
                pham = pha.copy()
                pham[nanmask] = 0
                pham = np.ma.masked_array(pham, mask=nanmask)
                pha = unwrap_phase(pham, seed=47)
                pha[nanmask] = np.nan
            else:
                pha = unwrap_phase(pha, seed=47)
            # remove 2PI offsets that might be present in the border phase
            border = np.concatenate((pha[0, :],
                                     pha[-1, :],
                                     pha[:, 0],
                                     pha[:, -1]))
            twopi = 2*np.pi
            minimum = divmod_neg(np.nanmin(border), twopi)[0]
            offset = minimum * twopi
            pha -= offset

        if bg_data is not None:
            bg_amp, bg_pha, _, _ = self._get_amp_pha(
                data=bg_data,
                which_data=which_data,
                bg_data=None,
                proc_phase=proc_phase)
        else:
            bg_amp = bg_pha = None

        return amp, pha, bg_amp, bg_pha

    @property
    def bg_amp(self):
        """background amplitude image"""
        return self._amp.bg

    @property
    def bg_pha(self):
        """background phase image"""
        return self._pha.bg

    @property
    def amp(self):
        """background-corrected amplitude image"""
        return self._amp.image

    @property
    def dtype(self):
        """dtype of the phase data array"""
        return self._pha.raw.dtype

    @property
    def field(self):
        """background-corrected complex field"""
        return self.amp * np.exp(1j * self.pha)

    @property
    def info(self):
        """list of tuples with QPImage metadata"""
        info = []
        # meta data
        meta = self.meta
        for key in meta:
            info.append((key, self.meta[key]))
        # background correction
        for imdat in [self._amp, self._pha]:
            info += imdat.info
        return info

    @property
    def meta(self):
        """dictionary with imaging meta data"""
        return MetaDict(self.h5.attrs)

    @property
    def pha(self):
        """background-corrected phase image"""
        return self._pha.image

    @property
    def raw_amp(self):
        """raw amplitude image"""
        return self._amp.raw

    @property
    def raw_pha(self):
        """raw phase image"""
        return self._pha.raw

    @property
    def shape(self):
        """size of image dimensions"""
        return self._pha.h5["raw"].shape

    def clear_bg(self, which_data=("amplitude", "phase"), keys="fit"):
        """Clear background correction

        Parameters
        ----------
        which_data: str or list of str
            From which type of data to remove the background
            information. The list contains either "amplitude",
            "phase", or both.
        keys: str or list of str
            Which type of background data to remove. One of:

            - "fit": the background data computed with
              :func:`qpimage.QPImage.compute_bg`
            - "data": the experimentally obtained background image
        """
        which_data = QPImage._conv_which_data(which_data)
        if isinstance(keys, str):
            # make sure keys is a list of strings
            keys = [keys]

        # Get image data for clearing
        imdats = []
        if "amplitude" in which_data:
            imdats.append(self._amp)
        if "phase" in which_data:
            imdats.append(self._pha)
        if not imdats:
            msg = "`which_data` must contain 'phase' or 'amplitude'!"
            raise ValueError(msg)
        # Perform clearing of backgrounds
        for imdat in imdats:
            for key in keys:
                imdat.del_bg(key)

    def compute_bg(self, which_data="phase",
                   fit_offset="mean", fit_profile="tilt",
                   border_m=0, border_perc=0, border_px=0,
                   from_mask=None, ret_mask=False):
        """Compute background correction

        Parameters
        ----------
        which_data: str or list of str
            From which type of data to remove the background
            information. The list contains either "amplitude",
            "phase", or both.
        fit_profile: str
            The type of background profile to fit:

            - "offset": offset only
            - "poly2o": 2D 2nd order polynomial with mixed terms
            - "tilt": 2D linear tilt with offset (default)
        fit_offset: str
            The method for computing the profile offset

            - "fit": offset as fitting parameter
            - "gauss": center of a gaussian fit
            - "mean": simple average
            - "mode": mode (see `qpimage.bg_estimate.mode`)
        border_m: float
            Assume that a frame of `border_m` meters around the
            image is background. The value is converted to
            pixels and rounded.
        border_perc: float
            Assume that a frame of `border_perc` percent around
            the image is background. The value is converted to
            pixels and rounded. If the aspect ratio of the image
            is not one, then the average of the data's shape is
            used to compute the percentage in pixels.
        border_px: float
            Assume that a frame of `border_px` pixels around
            the image is background.
        from_mask: boolean np.ndarray or None
            Use a boolean array to define the background area.
            The boolean mask must have the same shape as the
            input data. `True` elements are used for background
            estimation.
        ret_mask: bool
            Return the boolean mask used to compute the background.

        Notes
        -----
        The `border_*` values are translated to pixel values and
        the largest pixel border is used to generate a mask
        image for background computation.

        If any of the `border_*` arguments are non-zero and
        `from_mask` is given, the intersection of the two
        is used, i.e. the positions where both, the frame
        mask and `from_mask`, are `True`.

        See Also
        --------
        qpimage.bg_estimate.estimate
        """
        which_data = QPImage._conv_which_data(which_data)
        # check validity
        if not ("amplitude" in which_data or
                "phase" in which_data):
            msg = "`which_data` must contain 'phase' or 'amplitude'!"
            raise ValueError(msg)
        # get border in px
        border_list = []
        if border_m:
            if border_m < 0:
                raise ValueError("`border_m` must be greater than zero!")
            border_list.append(border_m / self.meta["pixel size"])
        if border_perc:
            if border_perc < 0 or border_perc > 50:
                raise ValueError("`border_perc` must be in interval [0, 50]!")
            size = np.average(self.shape)
            border_list.append(size * border_perc / 100)
        if border_px:
            border_list.append(border_px)
        # get maximum border size
        if border_list:
            border_px = int(np.round(np.max(border_list)))
        elif from_mask is None:
            raise ValueError("Neither `from_mask` nor `border_*` given!")
        elif np.all(from_mask == 0):
            raise ValueError("`from_mask` must not be all-zero!")
        # Get affected image data
        imdat_list = []
        if "amplitude" in which_data:
            imdat_list.append(self._amp)
        if "phase" in which_data:
            imdat_list.append(self._pha)
        # Perform correction
        mask = None
        for imdat in imdat_list:
            mask = imdat.estimate_bg(fit_offset=fit_offset,
                                     fit_profile=fit_profile,
                                     border_px=border_px,
                                     from_mask=from_mask,
                                     ret_mask=ret_mask)
        # return phase mask (if possible) for user-convenience/testing
        return mask

    def copy(self, h5file=None):
        """Create a copy of the current instance

        This is done by recursively copying the underlying HDF5 data.

        Parameters
        ----------
        h5file: str, h5py.File, h5py.Group, or None
            see `QPImage.__init__`
        """
        h5 = copyh5(self.h5, h5file)
        return QPImage(h5file=h5, h5dtype=self.h5dtype)

    def refocus(self, distance, kernel="helmholtz", padding=True,
                h5file=None, h5mode="a", ret_refocus_iface=False,
                method=None):
        """Compute a numerically refocused QPImage

        Parameters
        ----------
        distance: float
            Focusing distance [m]
        kernel: str
            Refocusing method, one of ["helmholtz","fresnel"]
        padding: bool
            Whether to perform padding during refocusing.
            You may disable padding if your input image does not
            have any discontinuities at the border (i.e. you can
            tile your input image, and it would look good), otherwise
            you will experience ringing artifacts.
        h5file: str, h5py.Group, h5py.File, or None
            A path to an HDF5 data file where the QPImage is cached.
            If set to `None` (default), all data will be handled in
            memory using the "core" driver of the :mod:`h5py`'s
            :class:`h5py:File` class. If the file does not exist,
            it is created. If the file already exists, it is opened
            with the file mode defined by `hdf5_mode`. If this is
            an instance of h5py.Group or h5py.File, then this will
            be used to internally store all data.
        h5mode: str
            Valid file modes are (only applies if `h5file` is a path)

            - "r": Readonly, file must exist
            - "r+": Read/write, file must exist
            - "w": Create file, truncate if exists
            - "w-" or "x": Create file, fail if exists
            - "a": Read/write if exists, create otherwise (default)
        ret_refocus_iface: bool
            Whether or not to also return the `nrefocus.Refocus*` class
            used for refocusing; this might be useful if you wish to
            do quick manual refocusing or have to create a refocusing
            series.
        method: str
            deprecated, use `kernel` instead

        Returns
        -------
        qpi: qpimage.QPImage
            Refocused phase and amplitude data

        See Also
        --------
        - :mod:`nrefocus`: library used for numerical focusing
        - :ref:`nrefocus:sec_refocus_interface`: refocusing interface
        """
        if method is not None:
            warnings.warn("`method` is deprecated, use `kernel` instead!",
                          DeprecationWarning)
            if kernel is not None:
                kernel = method
        # `self.field` this is always computed anew, so we don't have to worry
        # about comparing against the same object.
        field = self.field
        if (self._refocuser is None
            or not np.all(self._refocuser.origin == field)
            or self._refocuser.kernel != kernel
                or self._refocuser.padding != padding):
            # We have a new field, so we have to create a new Refocus
            refocus_iface = nrefocus.get_best_interface()
            self._refocuser = refocus_iface(
                field=field,
                wavelength=self["wavelength"],
                pixel_size=self["pixel size"],
                medium_index=self["medium index"],
                distance=0,
                kernel=kernel,
                padding=padding
            )
        field2 = self._refocuser.propagate(distance)

        if "identifier" in self:
            ident = self["identifier"]
        else:
            ident = ""
        meta_data = self.meta
        meta_data["identifier"] = f"{ident}@{kernel[0]}{distance:.5e}m"

        qpi2 = QPImage(data=field2,
                       which_data="field",
                       meta_data=meta_data,
                       h5file=h5file,
                       h5mode=h5mode)
        if ret_refocus_iface:
            return qpi2, self._refocuser
        else:
            return qpi2

    def set_bg_data(self, bg_data, which_data=None, proc_phase=True):
        """Set background amplitude and phase data

        Parameters
        ----------
        bg_data: 2d ndarray (float or complex), list, QPImage, or `None`
            The background data (must be same type as `data`).
            If set to `None`, the background data is reset.
        which_data: str
            String or comma-separated list of strings indicating
            the order and type of input data. Valid values are
            "field", "phase", "phase,amplitude", or "phase,intensity",
            where the latter two require an indexable object for `bg_data`
            with the phase data as first element.
        proc_phase: bool
            Process the phase data. This includes phase unwrapping
            using :func:`skimage.restoration.unwrap_phase` and
            correcting for 2PI phase offsets (The offset is estimated
            from a 1px-wide border around the image).
        """
        if isinstance(bg_data, QPImage):
            if which_data is not None:
                msg = "`which_data` must not be set if `bg_data` is QPImage!"
                raise ValueError(msg)
            pha, amp = bg_data.pha, bg_data.amp
        elif bg_data is None:
            # Reset phase and amplitude
            amp, pha = None, None
        else:
            # Compute phase and amplitude from data and which_data
            amp, pha, _, _ = self._get_amp_pha(data=bg_data,
                                               which_data=which_data,
                                               proc_phase=proc_phase)
        # Set background data
        self._amp.set_bg(amp, key="data")
        self._pha.set_bg(pha, key="data")


def copyh5(inh5, outh5):
    """Recursively copy all HDF5 data from one group to another

    Data from links is copied.

    Parameters
    ----------
    inh5: str, h5py.File, or h5py.Group
        The input HDF5 data. This can be either a file name or
        an HDF5 object.
    outh5: str, h5py.File, h5py.Group, or None
        The output HDF5 data. This can be either a file name or
        an HDF5 object. If set to `None`, a new HDF5 object is
        created in memory.

    Notes
    -----
    All data in outh5 are overridden by the inh5 data.
    """
    if not isinstance(inh5, h5py.Group):
        inh5 = h5py.File(inh5, mode="r")
    if outh5 is None:
        # create file in memory
        h5kwargs = {"name": "qpimage{}.h5".format(QPImage._instances),
                    "driver": "core",
                    "backing_store": False,
                    "mode": "w"}
        outh5 = h5py.File(**h5kwargs)
        return_h5obj = True
        QPImage._instances += 1
    elif not isinstance(outh5, h5py.Group):
        # create new file
        outh5 = h5py.File(outh5, mode="w")
        return_h5obj = False
    else:
        return_h5obj = True
    # begin iteration
    for key in inh5:
        if key in outh5:
            del outh5[key]
        if isinstance(inh5[key], h5py.Group):
            outh5.create_group(key)
            copyh5(inh5[key], outh5[key])
        else:
            dset = write_image_dataset(group=outh5,
                                       key=key,
                                       data=inh5[key][:],
                                       h5dtype=inh5[key].dtype)
            dset.attrs.update(inh5[key].attrs)
    outh5.attrs.update(inh5.attrs)
    if return_h5obj:
        # in-memory or previously created instance of h5py.File
        return outh5
    else:
        # properly close the file and return its name
        fn = outh5.filename
        outh5.flush()
        outh5.close()
        return fn


def divmod_neg(a, b):
    """Return divmod with the closest result to zero"""
    q, r = divmod(a, b)
    # make sure r is close to zero
    sr = np.sign(r)
    if np.abs(r) > b/2:
        q += sr
        r -= b * sr
    return q, r



import pathlib

import qpimage.integrity_check


def test_tilt_from_h5file():
    """"
    The data for this test was created using:


    ```
    import numpy as np
    import qpimage

    size = 50
    # background phase image with a tilt
    bg = np.repeat(np.linspace(0, 1, size), size).reshape(size, size)
    bg = .6 * bg - .8 * bg.transpose() + .2
    # phase image with random noise
    rsobj = np.random.RandomState(47)
    phase = rsobj.rand(size, size) - .5 + bg

    # create QPImage instance
    with qpimage.QPImage(data=phase,
                         which_data="phase",
                         h5file="bg_tilt.h5") as qpi:
        qpi.compute_bg(which_data="phase",  # correct phase image
                       fit_offset="fit",  # use bg offset from tilt fit
                       fit_profile="tilt",  # perform 2D tilt fit
                       border_px=5,  # use 5 px border around image
                       )
    ```
    """
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpimage.integrity_check.check(h5file, checks=["background"])


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Filter choices for interferometric imaging

There are several parameters that influence the quality of
phase and amplitude data retrieved from data recorded via
interferometric techniques. This example demonstrates the
advantages and disadvantages of three filtering stragies
in qpimage. For more information, please have a look at
the :ref:`qpretrieve <qpretrieve:index>` library.

Several observations can be made:

- There appears to be a "bleed-through" of phase data into the
  amplitude data.
- A (sharp) disk filter introduces ringing artifacts in the amplitude
  and phase images.
- A smooth disk filter does not lead to such artifacts, but a dark
  halo is introduced around the coins in the amplitude image.
- The amplitude reconstruction with the gaussian filter does not
  exhibit the dark halo but, due to blurring, reveals less details.

To correctly interpret the data shown, please note that:

- This is a simulated hologram with *no* central band. For real
  data, the "filter_size" parameter also affects the reconstruction
  quality. Contributions from the central band can lead to strong
  artifacts. A balance between high resolution (large filter size)
  and small contributions from the central band (small filter size)
  usually has to be found.
- It is not trivial to compare a gaussian filter with a disk filter
  in terms of filter size (sigma vs. radius).
  The gaussian filter takes into account larger frequencies and
  suppresses low frequencies. In qpimage, the actual gaussian filter
  size is chosen such that the resolution approximately matches that
  of the disk filter with a corresponding radius. In general however,
  the filter size parameter has to be examined when comparing the two.
- There is an inherent loss of information (resolution) in the
  holographic reconstruction process. The side band is isolated with a
  low-pass filter in Fourier space. The size and shape of this filter
  determine the resolution of the phase and amplitude images. As a
  result, the level of detail of all reconstructions shown is lower
  than that of the original images.
"""
import matplotlib.pylab as plt
import numpy as np
import qpimage
from skimage import color, data

# image of a galaxy recorded with the Hubble telescope
img1 = color.rgb2gray(data.hubble_deep_field())[354:504, 70:220]
# image of a coin
img2 = data.coins()[150:300, 70:220]

pha = img1/img1.max() * 2 * np.pi
amp = img2/img2.mean()

# create a hologram
x, y = np.mgrid[0:150, 0:150]
hologram = 2 * amp * np.cos(-2 * (x + y) + pha)

filters = ["disk", "smooth disk", "gauss"]
qpis = []

for filter_name in filters:
    qpi = qpimage.QPImage(data=hologram,
                          which_data="raw-oah",
                          qpretrieve_kw={"filter_size": .5,
                                         "filter_name": filter_name})
    qpis.append(qpi)

fig = plt.figure(figsize=(8, 16))

phakw = {"interpolation": "bicubic",
         "cmap": "viridis",
         "vmin": pha.min(),
         "vmax": pha.max(),
         }

ampkw = {"interpolation": "bicubic",
         "cmap": "gray",
         "vmin": amp.min(),
         "vmax": amp.max()
         }

numrows = len(filters) + 1

plt.subplot(numrows, 2, 1, title="original phase")
plt.imshow(pha, **phakw)

ax2 = plt.subplot(numrows, 2, 2, title="original amplitude")
plt.imshow(amp, **ampkw)

for ii in range(len(filters)):
    # phase
    plt.subplot(numrows, 2, 2*ii+3, title=filters[ii]+" filter")
    plt.imshow(qpis[ii].pha, **phakw)
    # amplitude
    plt.subplot(numrows, 2, 2*ii+4, title=filters[ii]+" filter")
    plt.imshow(qpis[ii].amp, **ampkw)

# disable axes
for ax in fig.get_axes():
    ax.axis("off")

plt.tight_layout()
plt.show()



import abc
import numbers
import warnings

import h5py
import numpy as np

from . import bg_estimate

#: default HDF5 compression keyword arguments
COMPRESSION = {"compression": "gzip",
               "compression_opts": 9,
               }

#: valid background data identifiers
VALID_BG_KEYS = ["data",
                 "fit",
                 ]


class ImageData(object):
    """Base class for image management

    See Also
    --------
    Amplitude: ImageData with amplitude background correction
    Phase: ImageData with phase background correction
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, h5, h5dtype="float32"):
        """
        Parameters
        ----------
        h5: h5py.Group
            HDF5 group where all data is kept

        h5dtype: str
            The datatype in which to store the image data. The default
            is "float32" which is sufficient for 2D image analysis and
            consumes only half the disk space of the numpy default
            "float64".
        """
        self.h5dtype = np.dtype(h5dtype)
        self.h5 = h5
        if "bg_data" not in self.h5:
            self.h5.create_group("bg_data")

    def __repr__(self):
        name = self.__class__.__name__
        rep = "{name} image, {x}x{y}px".format(name=name,
                                               x=self.raw.shape[0],
                                               y=self.raw.shape[1],
                                               )
        return rep

    def __setitem__(self, key, value):
        """Image data setter

        If `value` is None, then `key` is removed from `self.h5`.
        The datatype `self.h5dtype` is used, unless the input
        array is boolean.
        """
        if value is None:
            if key in self.h5:
                del self.h5[key]
        else:
            if value.dtype == np.dtype("bool"):
                h5dtype = "bool"
            else:
                h5dtype = self.h5dtype
            write_image_dataset(group=self.h5,
                                key=key,
                                data=value,
                                h5dtype=h5dtype)

    @abc.abstractmethod
    def _bg_combine(self, *bgs):
        """Combine several background images"""

    @abc.abstractmethod
    def _bg_correct(self, raw, bg):
        """Remove `bg` from `raw` image data"""

    @property
    def bg(self):
        """combined background image data"""
        return self._bg_combine(self.h5["bg_data"].values())

    @property
    def image(self):
        """background corrected image data"""
        return self._bg_correct(self.raw, self.bg)

    @property
    def info(self):
        """list of background correction parameters"""
        info = []
        name = self.__class__.__name__.lower()
        # get bg information
        for key in VALID_BG_KEYS:
            if key in self.h5["bg_data"]:
                attrs = self.h5["bg_data"][key].attrs
                for akey in attrs:
                    atr = attrs[akey]
                    var = "{} background {}".format(name, akey)
                    info.append((var, atr))
        if "fit" in self.h5["bg_data"]:
            # mask background
            var_mask = "{} background from mask".format(name)
            if ("estimate_bg_from_mask" in self.h5
                    and self.h5["estimate_bg_from_mask"] is not None):
                # bg was computed from mask image
                info.append((var_mask, True))
            elif ("estimate_bg_from_binary" in self.h5
                    and self.h5["estimate_bg_from_binary"] is not None):
                # bg was computed from mask image (old notation)
                warnings.warn("Old file format detected!", DeprecationWarning)
                info.append((var_mask, True))
            else:
                info.append((var_mask, False))
        return info

    @property
    def raw(self):
        """raw (uncorrected) image data"""
        return self.h5["raw"][:]

    def del_bg(self, key):
        """Remove the background image data

        Parameters
        ----------
        key: str
            One of :const:`VALID_BG_KEYS`
        """
        if key not in VALID_BG_KEYS:
            raise ValueError("Invalid bg key: {}".format(key))
        if key in self.h5["bg_data"]:
            del self.h5["bg_data"][key]
        else:
            msg = "No bg data to clear for '{}' in {}.".format(key, self)
            warnings.warn(msg)

    def estimate_bg(self, fit_offset="mean", fit_profile="tilt",
                    border_px=0, from_mask=None, ret_mask=False):
        """Estimate image background

        Parameters
        ----------
        fit_profile: str
            The type of background profile to fit:

            - "offset": offset only
            - "poly2o": 2D 2nd order polynomial with mixed terms
            - "tilt": 2D linear tilt with offset (default)
        fit_offset: str
            The method for computing the profile offset

            - "fit": offset as fitting parameter
            - "gauss": center of a gaussian fit
            - "mean": simple average
            - "mode": mode (see `qpimage.bg_estimate.mode`)
        border_px: float
            Assume that a frame of `border_px` pixels around
            the image is background.
        from_mask: boolean np.ndarray or None
            Use a boolean array to define the background area.
            The mask image must have the same shape as the
            input data.`True` elements are used for background
            estimation.
        ret_mask: bool
            Return the mask image used to compute the background.

        Notes
        -----
        If both `border_px` and `from_mask` are given, the
        intersection of the two resulting mask images is used.

        The arguments passed to this method are stored in the
        HDF5 file `self.h5` and are used for optional integrity
        checking using `qpimage.integrity_check.check`.

        See Also
        --------
        qpimage.bg_estimate.estimate
        """
        # remove existing bg before accessing imdat.image
        self.set_bg(bg=None, key="fit")
        # compute bg
        bgimage, mask = bg_estimate.estimate(data=self.image,
                                             fit_offset=fit_offset,
                                             fit_profile=fit_profile,
                                             border_px=border_px,
                                             from_mask=from_mask,
                                             ret_mask=True)
        attrs = {"fit_offset": fit_offset,
                 "fit_profile": fit_profile,
                 "border_px": border_px}
        self.set_bg(bg=bgimage, key="fit", attrs=attrs)
        # save `from_mask` separately (arrays vs. h5 attributes)
        # (if `from_mask` is `None`, this will remove the array)
        self["estimate_bg_from_mask"] = from_mask
        # return mask image
        if ret_mask:
            return mask

    def get_bg(self, key=None, ret_attrs=False):
        """Get the background data

        Parameters
        ----------
        key: None or str
            A user-defined key that identifies the background data.
            Examples are "data" for experimental data, or "fit"
            for an estimated background correction
            (see :const:`VALID_BG_KEYS`). If set to `None`,
            returns the combined background image (:const:`ImageData.bg`).
        ret_attrs: bool
            Also returns the attributes of the background data.
        """
        if key is None:
            if ret_attrs:
                raise ValueError("No attributes for combined background!")
            return self.bg
        else:
            if key not in VALID_BG_KEYS:
                raise ValueError("Invalid bg key: {}".format(key))
            if key in self.h5["bg_data"]:
                data = self.h5["bg_data"][key][:]
                if ret_attrs:
                    attrs = dict(self.h5["bg_data"][key].attrs)
                    # remove keys for image visualization in hdf5 files
                    for h5k in ["CLASS", "IMAGE_VERSION", "IMAGE_SUBCLASS"]:
                        if h5k in attrs:
                            attrs.pop(h5k)
                    ret = (data, attrs)
                else:
                    ret = data
            else:
                raise KeyError("No background data for {}!".format(key))
            return ret

    def set_bg(self, bg, key="data", attrs=None):
        """Set the background data

        Parameters
        ----------
        bg: numbers.Real, 2d ndarray, ImageData, or h5py.Dataset
            The background data. If `bg` is an `h5py.Dataset` object,
            it must exist in the same HDF5 file (a hard link is created).
            If set to `None`, the data will be removed.
        key: str
            One of :const:`VALID_BG_KEYS`)
        attrs: dict
            List of background attributes

        See Also
        --------
        del_bg: removing background data
        """
        if attrs is None:
            attrs = {}
        if key not in VALID_BG_KEYS:
            raise ValueError("Invalid bg key: {}".format(key))
        # remove previous background key
        if key in self.h5["bg_data"]:
            del self.h5["bg_data"][key]
        # set background
        if isinstance(bg, (numbers.Real, np.ndarray)):
            dset = write_image_dataset(group=self.h5["bg_data"],
                                       key=key,
                                       data=bg,
                                       h5dtype=self.h5dtype)
            for kw in attrs:
                dset.attrs[kw] = attrs[kw]
        elif isinstance(bg, h5py.Dataset):
            # Create a hard link
            # (This functionality was intended for saving memory when storing
            # large QPSeries with universal background data, i.e. when using
            # `QPSeries.add_qpimage` with the `bg_from_idx` keyword.)
            self.h5["bg_data"][key] = bg
        elif bg is not None:
            msg = "Unknown background data type: {}".format(bg)
            raise ValueError(msg)


class Amplitude(ImageData):
    """Dedicated class for amplitude image data

    For amplitude image data, background correction is defined
    by dividing the raw image by the background image.
    """

    def _bg_combine(self, bgs):
        """Combine several background amplitude images"""
        out = np.ones(self.h5["raw"].shape, dtype=float)
        # bg is an h5py.DataSet
        for bg in bgs:
            out *= bg[:]
        return out

    def _bg_correct(self, raw, bg):
        """Remove background from raw amplitude image"""
        return raw / bg


class Phase(ImageData):
    """Dedicated class for phase image data

    For phase image data, background correction is defined
    by subtracting the background image from the raw image.
    """

    def _bg_combine(self, bgs):
        """Combine several background phase images"""
        out = np.zeros(self.h5["raw"].shape, dtype=float)
        for bg in bgs:
            # bg is an h5py.DataSet
            out += bg[:]
        return out

    def _bg_correct(self, raw, bg):
        """Remove background from raw phase image"""
        return raw - bg


def write_image_dataset(group, key, data, h5dtype=None):
    """Write an image to an HDF5 group as a dataset

    This convenience function sets all attributes such that the image
    can be visualized with HDFView, sets the compression and fletcher32
    filters, and sets the chunk size to the image shape.

    Parameters
    ----------
    group: h5py.Group
        HDF5 group to store data to
    key: str
        Dataset identifier
    data: np.ndarray of shape (M,N)
        Image data to store
    h5dtype: str or dtype
        The datatype in which to store the image data. The default
        is the datatype of `data`.

    Returns
    -------
    dataset: h5py.Dataset
        The created HDF5 dataset object
    """
    if h5dtype is None:
        h5dtype = data.dtype
    if key in group:
        del group[key]
    if group.file.driver == "core":
        kwargs = {}
    else:
        kwargs = {"fletcher32": True,
                  "chunks": data.shape}
        kwargs.update(COMPRESSION)

    dset = group.create_dataset(key,
                                data=data.astype(h5dtype),
                                **kwargs)
    if group.file.driver != "core":
        # HDFView 3.1.1 supports visualizing float-arrays as grayscale image
        dset.attrs.create('CLASS', np.string_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
    return dset



"""Display changelog with links to GitHub issues

Usage
-----
The directive

   .. include_changelog:: ../CHANGELOG

adds the content of the changelog file into the current document.
References to GitHub issues are identified as "(#XY)" (with parentheses
and hash) and a link is inserted

 https://github.com/RI-imaging/{PROJECT}/issues/{XY}

where PROJECT ist the `project` variable defined in conf.py.
"""
import io
import re

from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes


class IncludeDirective(Directive):
    required_arguments = 1
    optional_arguments = 0

    def run(self):
        full_path = self.arguments[0]
        project = self.state.document.settings.env.config.github_project

        def insert_github_link(reobj):
            line = reobj.string
            instr = line[reobj.start():reobj.end()]
            issue = instr.strip("#()")
            link = "https://github.com/{}/issues/".format(project)
            rstlink = "(`#{issue} <{link}{issue}>`_)".format(issue=issue,
                                                             link=link)
            return rstlink

        with io.open(full_path, "r") as myfile:
            text = myfile.readlines()

        rst = []
        for line in text:
            line = line.strip("\n")
            if line.startswith("  ") and line.strip().startswith("-"):
                # list in list:
                rst.append("")
            if not line.startswith(" "):
                rst.append("")
                line = "version " + line
                rst.append(line)
                rst.append("-"*len(line))
            elif not line.strip():
                rst.append(line)
            else:
                line = re.sub(r"\(#[0-9]*\)", insert_github_link, line)
                rst.append(line)

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


def setup(app):
    app.add_config_value('github_project', "user/project", 'html')
    app.add_directive('include_changelog', IncludeDirective)
    return {'version': '0.1'}   # identifies the version of our extension



"""Background image offset correction

This example illustrates the different background offset
correction methods implemented in qpimage. The phase image
data contains two gaussian noise distributions for which
these methods yield different background phase offsets.
"""
import matplotlib.pylab as plt
import numpy as np
import qpimage

size = 200  # the size of the image
bg = 2.5  # the center of the background phase distribution
scale = .1  # the spread of the background phase distribution

# compute random phase data
rsobj = np.random.RandomState(42)
data = rsobj.normal(loc=bg, scale=scale, size=size**2)
# Add a second distribution `data2` at random positions `idx`,
# such that there is no pure gaussian distribution.
# (otherwise 'mean' and 'gaussian' cannot be distinguished)
data2 = rsobj.normal(loc=bg*1.1, scale=scale, size=size**2//2)
idx = rsobj.choice(data.size, data.size//2)
data[idx] = data2
# reshape `data` to get a 2D array
data = data.reshape(size, size)

qpi = qpimage.QPImage(data=data, which_data="phase")

cpkw = {"which_data": "phase",  # correct the input phase data
        "fit_profile": "offset",  # perform offset correction only
        "border_px": 5,  # use a border of 5px of the input phase
        "ret_mask": True,  # return the mask image for visualization
        }

mask = qpi.compute_bg(fit_offset="mode", **cpkw)
bg_mode = np.mean(qpi.bg_pha[mask])

qpi.compute_bg(fit_offset="mean", **cpkw)
bg_mean = np.mean(qpi.bg_pha[mask])

qpi.compute_bg(fit_offset="gauss", **cpkw)
bg_gauss = np.mean(qpi.bg_pha[mask])

bg_data = (qpi.pha + qpi.bg_pha)[mask]
# compute histogram
nbins = int(np.ceil(np.sqrt(bg_data.size)))
mind, maxd = bg_data.min(), bg_data.max()
histo = np.histogram(bg_data, nbins, density=True, range=(mind, maxd))
dx = abs(histo[1][1] - histo[1][2]) / 2
hx = histo[1][1:] - dx
hy = histo[0]

# plot the properties of `qpi`
plt.figure(figsize=(8, 4))

ax1 = plt.subplot(121, title="input phase")
map1 = plt.imshow(data)
plt.colorbar(map1, ax=ax1, fraction=.046, pad=0.04)


t2 = "{}px border histogram with {} bins".format(cpkw["border_px"], nbins)
plt.subplot(122, title=t2)
plt.plot(hx, hy, label="histogram", color="gray")
plt.axvline(bg_mode, 0, 1, label="mode", color="red")
plt.axvline(bg_mean, 0, 1, label="mean", color="green")
plt.axvline(bg_gauss, 0, 1, label="gauss", color="orange")
plt.legend()

plt.tight_layout()
plt.show()



import numpy as np

import qpimage  # noqa: E402


def test_clear_bg():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)

    bg_amp = np.linspace(.93, 1.02, size**2).reshape(size, size)
    bg_pha = pha * .5

    qpi = qpimage.QPImage(data=(pha, amp),
                          bg_data=(bg_pha, bg_amp),
                          which_data="phase,amplitude",
                          h5dtype="float64")
    # amplitude
    assert np.all(qpi.bg_amp == bg_amp)
    qpi.clear_bg(which_data="amplitude", keys="data")
    assert np.allclose(qpi.bg_amp, 1)
    qpi.compute_bg(which_data="amplitude",
                   fit_offset="fit",
                   fit_profile="tilt",
                   border_px=5)
    assert not np.allclose(qpi.bg_amp, 1)
    qpi.clear_bg(which_data="amplitude", keys="fit")
    assert np.allclose(qpi.bg_amp, 1)
    # phase
    assert np.all(qpi.bg_pha == bg_pha)
    qpi.clear_bg(which_data="phase", keys="data")


def test_clear_bg_error():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    qpi = qpimage.QPImage(pha, which_data="phase")
    try:
        qpi.clear_bg(which_data="gretel")
    except ValueError:
        pass
    else:
        assert False, "clear bg of invalid which_data should not work"


def test_comput_bg_error():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    qpi = qpimage.QPImage(pha, which_data="phase",
                          meta_data={"pixel size": .1})
    try:
        qpi.compute_bg(which_data="hänsel", border_px=4)
    except ValueError:
        pass
    else:
        assert False, "invalid which_data used for compute_bg"
    try:
        qpi.compute_bg(which_data="phase")
    except ValueError:
        pass
    else:
        assert False, "computation of bg from undefined mask"
    try:
        qpi.compute_bg(which_data="phase", border_m=-1)
    except ValueError:
        pass
    else:
        assert False, "negative border not allowed"
    try:
        qpi.compute_bg(which_data="phase", border_perc=60)
    except ValueError:
        pass
    else:
        assert False, "more than 50 percent border not allowed"
    try:
        qpi.compute_bg(which_data="phase", border_perc=-10)
    except ValueError:
        pass
    else:
        assert False, "negative percent border not allowed"
    try:
        qpi.compute_bg(which_data="phase",
                       from_mask=np.zeros((size, size), dtype=bool))
    except ValueError:
        pass
    else:
        assert False, "all-zero mask array does not work"


def test_get_amp_pha():
    # make sure no errors when printing repr
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)
    field = amp * np.exp(1j * pha)
    inten = amp**2

    qpi1 = qpimage.QPImage(pha, which_data="phase")
    qpi2 = qpimage.QPImage([pha, amp], which_data="phase,amplitude")
    qpi3 = qpimage.QPImage([pha, inten], which_data="phase,intensity")
    qpi4 = qpimage.QPImage(field, which_data="field")

    # test phases
    assert np.allclose(qpi1.pha, qpi2.pha)
    assert np.allclose(qpi1.pha, qpi3.pha)
    assert np.allclose(qpi1.pha, qpi4.pha)

    # test amplitudes
    assert np.allclose(qpi1.amp, 1)
    assert np.allclose(qpi2.amp, qpi3.amp)
    assert np.allclose(qpi2.amp, qpi4.amp)


def test_get_amp_pha_invalid():
    size = 50
    data = np.zeros((size, size))
    try:
        qpimage.QPImage(data, which_data="jean-luc")
    except ValueError:
        pass
    else:
        assert False, "invalid which_data should not work"


def test_get_amp_pha_holo():
    # create fake hologram
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    kx = -.6
    ky = -.4
    image = np.sin(kx * x + ky * y)
    qpi = qpimage.QPImage(image, which_data="raw-oah")
    qpi.compute_bg(which_data="phase",
                   fit_offset="fit",
                   fit_profile="tilt",
                   border_px=5)
    assert np.abs(qpi.pha).max() < .1


def test_get_amp_pha_nan():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    phanan = pha.copy()
    phanan[:4] = np.nan
    qpi = qpimage.QPImage(phanan, which_data="phase")
    assert np.allclose(qpi.pha, phanan, equal_nan=True)


def test_get_amp_pha_raw():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)

    qpi = qpimage.QPImage([pha, amp], which_data="phase,amplitude")
    qpi.set_bg_data(bg_data=(pha * 5 + 1, amp + .1),
                    which_data=["phase", "amplitude"])

    # test phases
    assert np.allclose(pha, qpi.raw_pha)
    assert np.allclose(amp, qpi.raw_amp)


def test_get_amp_pha_unrwap():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    phastep = pha.copy()
    phastep[5:] += 2 * np.pi
    # phastep will be unrwapped
    qpi = qpimage.QPImage(phastep, which_data="phase")
    assert np.allclose(qpi.pha, pha)


def test_get_amp_pha_unrwap_disabled():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    phastep = pha.copy()
    phastep[5:] += 2 * np.pi
    # phastep will not be unrwapped
    qpi = qpimage.QPImage(phastep, which_data="phase", proc_phase=False)
    assert np.allclose(qpi.pha, phastep)


def test_properties():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)

    bg_amp = np.linspace(.93, 1.02, size**2).reshape(size, size)
    bg_pha = pha * .5

    qpi = qpimage.QPImage((pha, amp), bg_data=(bg_pha, bg_amp),
                          which_data="phase,amplitude",
                          h5dtype="float64")

    assert np.all(qpi.bg_amp == bg_amp)
    assert np.all(qpi.bg_pha == bg_pha)

    assert np.iscomplexobj(qpi.field)
    fval = -0.46418608511978926 + 0.91376822116221712j
    assert np.allclose(qpi.field[20, 20], fval)

    assert qpi.shape == (size, size)

    qpi.compute_bg(which_data=["phase", "amplitude"],
                   fit_offset="fit",
                   fit_profile="tilt",
                   border_px=5)

    assert not np.all(qpi.bg_amp == bg_amp)
    assert not np.all(qpi.bg_pha == bg_pha)


def test_set_bg_data_qpimage():
    size = 20
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)

    bg_amp = np.linspace(.93, 1.02, size**2).reshape(size, size)
    bg_pha = pha * .5

    qpi = qpimage.QPImage(data=(pha, amp),
                          which_data="phase,amplitude")
    bg_qpi = qpimage.QPImage(data=(bg_pha, bg_amp),
                             which_data="phase,amplitude")
    qpi.set_bg_data(bg_qpi)
    assert np.allclose(qpi.pha, pha - bg_pha)

    # test wrong kwarg
    try:
        qpi.set_bg_data(bg_qpi, which_data="phase,amplitude")
    except ValueError:
        pass
    else:
        assert False, "which_data is invalid argument when data is QPImage"


def test_slice():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)
    bg_pha = np.zeros_like(pha) + np.linspace(-.1, .1, size).reshape(-1, 1)
    bg_amp = np.zeros_like(pha) + np.linspace(1.1, .99, size).reshape(-1, 1)

    qpi = qpimage.QPImage(data=[pha, amp],
                          bg_data=[bg_pha, bg_amp],
                          which_data="phase,amplitude",
                          meta_data={"wavelength": 500e-9})
    x = 25
    y = 10
    x_size = 25
    y_size = 5
    qpic = qpi[x:x + x_size, y:y + y_size]
    # simple sanity checks
    assert qpic.shape == (x_size, y_size)
    # check bg_correction
    assert np.allclose(qpic.pha, (pha - bg_pha)[x:x + x_size, y:y + y_size])
    assert np.allclose(qpic.amp, (amp / bg_amp)[x:x + x_size, y:y + y_size])
    # check bg data preservation (new in 0.5.0)
    assert np.allclose(qpic.bg_pha, bg_pha[x:x + x_size, y:y + y_size])
    assert np.allclose(qpic.bg_amp, bg_amp[x:x + x_size, y:y + y_size])

    # slice along x
    qpic2 = qpi[x:x + x_size]
    assert qpic2.shape == (25, size)
    assert np.allclose(qpic2.pha, (pha - bg_pha)[x:x + x_size])
    assert np.allclose(qpic2.amp, (amp / bg_amp)[x:x + x_size])

    # index should not work
    try:
        qpi[0]
    except ValueError:
        pass
    else:
        assert False, "simple indexing not supported"

    # string returns meta
    assert qpi["wavelength"] == qpi.meta["wavelength"]


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""qpimage definition extraction

Usage
-----
Directives:

Table output of qpimage.meta.DATA_DEF

   .. qpimage_meta_table:: data

Table output of qpimage.meta.OTHER_DEF

   .. qpimage_meta_table:: other
"""
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

from qpimage import meta


class Base(Directive):
    required_arguments = 0
    optional_arguments = 0

    def generate_rst(self):
        pass

    def run(self):
        rst = self.generate_rst()

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


class MetaTable(Base):
    required_arguments = 1

    def generate_rst(self):
        which = self.arguments[0]
        if which == "data":
            ddict = meta.DATA_DEF
        elif which == "other":
            ddict = meta.OTHER_DEF
        else:
            raise ValueError("Expected argument 'data' or 'other', "
                             + "not '{}'.".format(which))
        rst = []

        keys = sorted(ddict.keys())

        rst.append(".. csv-table::")
        rst.append("    :header: key, description")
        rst.append("    :delim: tab")
        rst.append("")

        for kk in keys:
            rst.append("    {}\t {}".format(kk, ddict[kk]))

        rst.append("")

        return rst


def setup(app):
    app.add_directive('qpimage_meta_table', MetaTable)
    return {'version': '0.1'}   # identifies the version of our extension



import pathlib
import tempfile

import numpy as np

import qpimage


def test_h5file_confusion():
    # make up temporary series file
    seriesfile = tempfile.mktemp(suffix=".h5", prefix="test_qpimage_series_")
    try:
        # h5file should be specified with its corresponding parameter
        qpimage.QPSeries(seriesfile)
    except ValueError:
        pass
    else:
        assert False, "h5file must be given as kwarg!"


def test_getitem():
    size = 20
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)

    qpi1 = qpimage.QPImage(data=1.1 * pha,
                           which_data="phase")
    qpi2 = qpimage.QPImage(data=1.2 * pha,
                           which_data="phase")
    qpi3 = qpimage.QPImage(data=1.3 * pha,
                           which_data="phase")

    series = qpimage.QPSeries(qpimage_list=[qpi1, qpi2, qpi3])

    assert qpi1 != qpi2
    assert qpi1 != qpi3
    assert series[0] == qpi1
    assert series[1] == qpi2
    assert series[2] == qpi3
    assert series[-3] == qpi1
    assert series[-2] == qpi2
    assert series[-1] == qpi3

    try:
        series[-4]
    except ValueError:
        pass
    else:
        assert False, "Negative index exceeds size."


def test_getitem_identifier():
    size = 20
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)

    qpi1 = qpimage.QPImage(data=1.1 * pha,
                           which_data="phase",
                           meta_data={"identifier": "peter"})
    qpi2 = qpimage.QPImage(data=1.2 * pha,
                           which_data="phase",
                           meta_data={"identifier": "hans"})
    qpi3 = qpimage.QPImage(data=1.3 * pha,
                           which_data="phase",
                           meta_data={"identifier": "doe"})

    series = qpimage.QPSeries(qpimage_list=[qpi1, qpi2, qpi3])
    assert "peter" in series
    assert "peter_bad" not in series
    assert series["peter"] == qpi1
    assert series["peter"] != qpi2
    assert series["hans"] == qpi2
    assert series["doe"] == qpi3
    try:
        series["john"]
    except KeyError:
        pass
    else:
        assert False, "'john' is not in series"
    try:
        series.add_qpimage(qpi1)
    except ValueError:
        pass
    else:
        assert False, "Adding QPImage with same identifier should not work"


def test_identifier():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi = qpimage.QPImage(h5file=h5file, h5mode="r")
    series1 = qpimage.QPSeries(qpimage_list=[qpi, qpi, qpi],
                               identifier="test_identifier")
    assert series1.identifier == "test_identifier"

    series2 = qpimage.QPSeries(qpimage_list=[qpi, qpi, qpi])
    assert series2.identifier is None


def test_identifier_qpimage():
    size = 20
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)

    qpi1 = qpimage.QPImage(data=1.1 * pha,
                           which_data="phase")
    qpi2 = qpimage.QPImage(data=1.2 * pha,
                           which_data="phase")
    qpi3 = qpimage.QPImage(data=1.3 * pha,
                           which_data="phase")

    series = qpimage.QPSeries(qpimage_list=[qpi1, qpi2])
    series.add_qpimage(qpi=qpi3, identifier="hastalavista")
    assert series[2]["identifier"] == "hastalavista"


def test_iter():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi = qpimage.QPImage(h5file=h5file, h5mode="r")
    series = qpimage.QPSeries(qpimage_list=[qpi, qpi, qpi])

    for qpj in series:
        assert qpj == qpi


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Masked background image correction

This example illustrates background correction with qpimage using
a mask to exclude regions that do not contain background
information.

The phase image of a microgel bead (top left) has two artifacts; there
is a tilt-like phase profile added along the vertical axis and there is
a second microgel bead in close proximity to the center bead. A regular
phase tilt background correction using the image values around a frame
of five pixels (see "background_tilt.py" example) does not yield a flat
background, because the second bead is fitted into the background which
leads to a horizontal background phase profile (top right). By defining
a mask (bottom left image), the phase values of the second bead
can be excluded from the background tilt fit and a flat background
phase is achieved (bottom right).
"""
import matplotlib.pylab as plt
import numpy as np
import qpimage


# load the experimental data
input_phase = np.load("./data/phase_beads_close.npz")["phase"].astype(float)

# create QPImage instance
qpi = qpimage.QPImage(data=input_phase,
                      which_data="phase")

# background correction without mask
qpi.compute_bg(which_data="phase",
               fit_offset="fit",
               fit_profile="tilt",
               border_px=5,
               )
pha_nomask = qpi.pha

# educated guess for mask
mask = input_phase < input_phase.max() / 10

# background correction with mask
# (the intersection of `mask` and the 5px border is used for fitting)
qpi.compute_bg(which_data="phase",
               fit_offset="fit",
               fit_profile="tilt",
               border_px=5,
               from_mask=mask
               )
pha_mask = qpi.pha

# plot
fig = plt.figure(figsize=(8, 7))
plot_kw = {"vmin": -.1,
           "vmax": 1.5}

ax1 = plt.subplot(221, title="input phase")
map1 = ax1.imshow(input_phase, **plot_kw)
plt.colorbar(map1, ax=ax1, fraction=.044, pad=0.04)

ax2 = plt.subplot(222, title="tilt-corrected (no mask)")
map2 = ax2.imshow(pha_nomask, **plot_kw)
plt.colorbar(map2, ax=ax2, fraction=.044, pad=0.04)

ax3 = plt.subplot(223, title="mask")
map3 = ax3.imshow(1.*mask, cmap="gray_r")
plt.colorbar(map3, ax=ax3, fraction=.044, pad=0.04)

ax4 = plt.subplot(224, title="tilt-corrected (with mask)")
map4 = ax4.imshow(pha_mask, **plot_kw)
plt.colorbar(map4, ax=ax4, fraction=.044, pad=0.04)

# disable axes
[ax.axis("off") for ax in [ax1, ax2, ax3, ax3, ax4]]

plt.tight_layout(h_pad=0, w_pad=0)
plt.show()
